      program skyfmt

C***********************************************************************
C*                                                                     *
C*    This code reads a matrix stored in coordinate format             *
C*    and rewrites it in skyline fashion in a binary file.             *
C*                                                                     *
C***********************************************************************
c
c.... set limits .......................................................
c
c     ma     : maximum number of nonzeros
c     mb     : maximum block size          
c     mn     : maximum number of equations 
c     mblock : maximum number of blocks  
c
      integer   ma,mb,mblock,mn
c
      parameter (ma     =  135000)
      parameter (mb     =  800000)
      parameter (mn     =    1200)
      parameter (mblock =     100)
c
c.... work variables ...................................................
c
      integer   lbandw,nblock,ncfb,ncft,neq,non0
      integer   ia1,ia2,im1,im2,im3,im4
c
c.... arrays ...........................................................
c
      integer   m(ma+ma+mn+mblock*3)
      real      a(ma+mb)
c
C***********************************************************************
c
      neq    = 0
      non0   = 0
      ncft   = 0
      ncfb   = 0
      lbandw = 0
c
      write (*,2500) mn,ma,mb,mblock
c
      ia1 = 1
      ia2 = ia1 + ma
c
      im1 = 1
      im2 = im1 + ma
      im3 = im2 + ma
      im4 = im3 + mn
c
c.... reading matrix ...................................................
c
      call input  (a(ia1),m(im1),m(im2),ma,mn,neq,non0)
c
c.... numbering reordering .............................................
c
      call invert (m(im1),m(im2),m(im3),neq,non0)
c
c.... verifying storage area ...........................................
c
      call length (m(im1),m(im2),m(im3),lbandw,ncft,neq,non0)
c
c.... verifying blocking ...............................................
c
      call block  (m(im3),m(im4),mb,mblock,nblock,ncfb,neq)
c
c.... writing matrices .................................................
c
      call output (a(ia1),a(ia2),m(im1),m(im2),m(im3),m(im4),
     &            lbandw,mb,nblock,ncfb,ncft,neq,non0)
c
      write (*,2510) neq,non0,ncft,lbandw,nblock
c
 2500 format (/,'*********** SKYLINE TRANSFORMATION ***********'/
     &        /,'max number of equations .......... =',i10,
     &        /,'max number of entries (matrix) ... =',i10,
     &        /,'max number of entries (profile) .. =',i10,
     &        /,'max number of blocks ............. =',i10/)
 2510 format (/,'number of equations .............. =',i10,
     &        /,'number of entries (matrix) ....... =',i10,
     &        /,'number of entries (profile) ...... =',i10,
     &        /,'half bandwidth ................... =',i10,
     &        /,'number of blocks ................. =',i10)
c
      stop
      end
c
c==== subroutine input =================================================
c
      subroutine input (a,i,j,ma,mn,neq,non0)
c
      character name*30
      integer   i(*),j(*),k,ma,mn,neq,non0
      real      a(*)
      logical   iexist
c
c.... define file to be transformed ....................................
c
      write   (*,*)       'file to be transformed?'
      read    (*,'(a30)') name 
      inquire (file=name,exist=iexist)
      if (.not.iexist) then    
         write (*,'(a,a30)') 'FILE CANNOT BE OPENED :',name 
         stop
      end if
      open    (10,file=name)
      rewind  (10)
c
c.... read matrix ......................................................
c
      write (*,*) '* reading matrix *'
c
      do 010 k = 1,ma
         read  (10,*,end=020) i(k),j(k),a(k)
         neq  = max0(i(k),j(k))
         non0 = non0 + 1
  010 continue
         read  (10,*,end=020) i(ma),j(ma),a(ma)
         write (*,*) 'NOT ENOUGH WORK SPACE FOR READING MATRIX'
         stop
c
  020 continue
c
      if ( neq .gt. mn ) then
         write (*,*) 'NOT ENOUGH WORK SPACE FOR EQUATIONS'
         stop
      end if
c
      return
      end
c
c==== subroutine invert ================================================
c
      subroutine invert (i,j,perm,neq,non0)
c
      integer i(*),j(*),k,neq,non0,perm(*)
      logical iexist
c
      inquire (file='invp.dat',exist=iexist)
c
      if ( .not.iexist ) return
c
      write (*,*) '* permutation phase *' 
c
      open (13,file='invp.dat',access='sequential',form='formatted')
      read (13,*) (perm(k),k=1,neq)
c
      do 010 k = 1,non0
         i(k) = perm(i(k))
         j(k) = perm(j(k))
  010 continue
c
      return
      end
c
c==== subroutine lenght ================================================
c
      subroutine length (i,j,height,lbandw,ncft,neq,non0)
c
      integer height(*),i(*),ic,ir,j(*),k,lbandw,ncft,neq,non0
c
      write (*,*) '* verifying storage area *' 
c
      do 010 k = 1,neq
         height(k) = 0
  010 continue
c
      do 020 k = 1,non0
         ir  = min0(i(k),j(k))
         ic  = max0(i(k),j(k))
         height(ic) = max0(height(ic),ic-ir+1)
  020 continue
c
      do 030 k = 1,neq 
         ncft  = ncft + height(k)
         lbandw = max0(lbandw,height(k))
  030 continue
c
      return
      end
c
c==== subroutine block  ================================================
c
      subroutine block (height,ibl,mb,mblock,nblock,ncfb,neq)
c
      integer height(*),ibl(3,*),isom,k,lmin,mb,mblock,nblock,ncfb,neq
c
      write (*,*) '* verifying blocking *'
c
      isom  = 0
      lmin  = neq
      nblock = 1
      do 010 k = 1,neq
         isom = isom + height(k)
         if ( isom .gt. mb )   then
            if ( height(k) .gt. mb ) then
               write (*,*) 'BLOCK SIZE TOO SMALL'
               stop
            end if
            if ( nblock .eq. mblock ) then
               write (*,*) 'NUMBER OF BLOCKS EXCEEDED'
               stop
            end if
            ibl(2,nblock) = k - 1
            ibl(3,nblock) = lmin
            nblock = nblock + 1
            ibl(1,nblock) = k
            lmin = neq
            isom = height(k)
         end if
         lmin = min0(lmin,k-height(k)+1)
         height(k) = isom
  010 continue
      ibl(1,1) = 1
      ibl(2,nblock) = neq
      ibl(3,nblock) = lmin
      do 020 k = 1,nblock
         ncfb = max0(ncfb,height(ibl(2,k)))
  020 continue
      return
      end
c
c==== subroutine output ================================================
c
      subroutine output (a,b,i,j,height,ibl,lbandw,mb,
     &                  nblock,ncfb,ncft,neq,non0)
c
      integer height(*),i(*),ibl(3,*),ic,ir,j(*),jmax,jmin,
     &        k,l,lbandw,mb,nblock,ncfb,ncft,neq,non0
      real    a(*),b(*)
c
      write (*,*) '* writing information *'
c
      open  (11,file='info.bin',access='sequential',form='unformatted')
      write (11) lbandw,nblock,ncfb,ncft,neq
      write (11) (ibl(1,k),ibl(2,k),ibl(3,k),k=1,nblock) 
      write (11) (height(k),k=1,neq)
c
      open  (13,file='op_A.bin',access='sequential',form='unformatted')
c
      do 010 k = 1,nblock
         jmin = ibl(1,k)
         jmax = ibl(2,k)
         do 020 l = 1,mb
            b(l) = 0.0e0
  020    continue
         do 030 l = 1,non0
            ir = min0(i(l),j(l))
            ic = max0(i(l),j(l))
            if ( (jmin.le.ic) .and. (ic.le.jmax) ) then
               b(height(ic)-ic+ir) = a(l)
            end if
  030    continue
         write (13) (b(l),l=1,height(jmax))
  010 continue
c
      return
      end
